#-----------------------------------------------------------------------------#
# University of Milan: DAPS&CO 2021
# - Social Network Analysis   (Class 8)
# Instructor: Rob Fahey <robfahey@aoni.waseda.jp>
#-----------------------------------------------------------------------------#

Sys.setlocale("LC_ALL", 'en_GB.UTF-8')
Sys.setenv(LANG = "en_US.UTF-8")
setwd("~/Dropbox/Milan/2021 Social Network Analysis/Class 8 - Week 2 Wednesday/")   
# Set the above to the folder on your computer where you've saved the files for this class.



#-----------------------------------------------------------------------------#
# 0.  Reloading Yesterday's Data
#-----------------------------------------------------------------------------#

# Like any other R object, you can save an igraph network in an RData file.

library(igraph)
library(RColorBrewer)
load("maga_network.rdata")


# This is pretty much the last plot we ended up working on yesterday...
set.seed(500)
l <- layout_with_fr(maga_net_trim)
l <- norm_coords(l, ymin=-1, ymax=1, xmin=-1, xmax=1)
plot(maga_net_trim, 
     vertex.size = log(V(maga_net_trim)$degree)*1.2, 
     vertex.shape = "circle",
     vertex.color = brewer.pal(n=8, name="Set2")[V(maga_net_trim)$community],
     vertex.label.color = "black",
     vertex.label.family = "sans",
     vertex.label.cex = .6,
     vertex.frame.color = "Lightgrey",
     edge.arrow.mode = 0, 
     edge.color = adjustcolor("SkyBlue", alpha.f = .2),
     rescale = FALSE,
     layout = l*1.3)


#-----------------------------------------------------------------------------#
# 1.  Analysing Communities in the Network
#-----------------------------------------------------------------------------#


# We saw yesterday that the largest communities in our data were 3, 5 and 1 (the fourth-largest 
# community was dramatically smaller than these three very large groups). Let's look at some ways
# to analyse these communities and uncover their identifying factors.

focus_communities <- c(3, 5, 1)

# First, just to make things a little more clear for ourselves, we'll create a version of the 
# graph which only contains those communities, and take a look at them.

maga_net <- induced_subgraph(maga_net_trim, V(maga_net_trim)[community %in% focus_communities])
l <- layout_with_fr(maga_net)
plot(maga_net, 
     vertex.size = log(V(maga_net)$degree)*1.2, 
     vertex.shape = "circle",
     vertex.color = brewer.pal(n=8, name="Set2")[V(maga_net)$community],
     vertex.label.color = "black",
     vertex.label.family = "sans",
     vertex.label.cex = .6,
     vertex.frame.color = "Lightgrey",
     edge.arrow.mode = 0, 
     edge.color = adjustcolor("SkyBlue", alpha.f = .2),
     layout = l)

# Does this kind of layout look familiar? You might recall the famous graph of Republican and 
# Democrat blogs from 2004, which really kicked of the whole field of online political network
# analysis. It had a very similar shape; two dense clusters with only minimal connections 
# between them.

# This kind of network shape is suggestive of strong polarisation. Can we somehow confirm that
# that's what we're seeing in this network data?



#-----------------------------------------------------------------------------#
# 2.  Using Tweet Text to Identify Clusters
#-----------------------------------------------------------------------------#

# Perhaps the most straightforward way to identify each community in a retweet network is to
# look at which tweets are actually being distributed around that community.

# As previously, we need to identify the most important accounts in each community. I'm using
# eigenvector centrality to do this, and to simplify things a little, I'm going to write a 
# quick R function to return the three most central accounts for any given community.

community_eigen <- function(a_graph, x) {
  this_subcom <- induced_subgraph(a_graph, V(a_graph)$community == x)
  top_names <- names(head(sort(eigen_centrality(this_subcom)$vector, decreasing = TRUE), 3))
  return(top_names)
}

community_eigen(maga_net, 3)

# First, let's tidy up the maga_tweets data a little bit - there's a lot of data there we don't need,
# which can make it more difficult than necessary to work with this dataframe.

# I'm only going to keep the columns we actually need, and will also get rid of the non-retweet data.
maga_tweets <- maga_tweets %>% 
  dplyr::select(created_at, text, is_retweet, screen_name, description, retweet_screen_name, retweet_description) %>%
  dplyr::filter(is_retweet == "TRUE")

# Okay, now let's write a quick function to return the tweets sent by the most central accounts in each cluster.
# This function is pretty simple; it filters the maga_tweets frame to only contain retweets originally sent by
# the top accounts in the specified cluster (using the community_eigen() function we created a moment ago to
# calculate that), then it uses the distinct() function to remove any repetition.

top_tweets <- function(a_graph, x) {
  tweets <- maga_tweets %>% 
    dplyr::filter(retweet_screen_name %in% community_eigen(a_graph, x)) %>%
    dplyr::select(retweet_screen_name, text) %>%
    dplyr::distinct()
  return(tweets)
}

# Let's try this on each of our clusters. Stop after each one and see what you can gather about that cluster from
# the tweets presented in the results.
top_tweets(maga_net, 3)

top_tweets(maga_net, 5)

top_tweets(maga_net, 1)

# Of course, you don't have to stop at visual observation of the tweets! You could also take these tables of data
# and pass them through to a package like Quanteda to perform a more rigorous text analysis. In a situation where
# you're dealing with a lot more data (potentially hundreds or thousands of tweets for each community), this
# would probably be essential.

#-----------------------------------------------------------------------------#
# 3.  Text Mining User Descriptions for Cluster Identification
#-----------------------------------------------------------------------------#

# Let's look at another way to analyse these communities: by looking at the profiles of the people who make up
# each community.

# For this approach, we're going to look at the "description" field for each of the users in each community.

# We're not restricting ourselves to only the top three users - we're going to look at the description text for
# literally everyone in the community. Therefore, it's going to be impossible to do this without doing some
# simple text mining.

# I'm sure you have Quanteda installed already, but if not:
install.packages("quanteda")

library(quanteda)

# We can get the names of the members of each community pretty easily, by filtering the nodes that match
# each community:
names(V(maga_net)[community == 3])

# Now let's create a function that's going to find the description text of those people, then use quanteda
# to create a word-cloud of the most significant words in each cluster.

description_cloud <- function(a_graph, x) {
  maga_tweets %>%
    dplyr::filter(screen_name %in% names(V(a_graph)[community == x])) %>%
    dplyr::select(screen_name, description) %>% 
    dplyr::distinct() %>%
    corpus(docid_field = "screen_name", text_field = "description") %>%
    tokens(remove_punct = TRUE, remove_symbols = TRUE) %>%
    tokens_remove(stopwords("english")) %>%
    dfm() %>%
    textplot_wordcloud(max_words = 20)
}

description_cloud(maga_net, 3)

description_cloud(maga_net, 5)

description_cloud(maga_net, 1)

# A word cloud is the most basic kind of analysis we could do here; naturally, you could also do any sort
# of complex text analysis (sentiment analysis, topic modelling etc.) that you thought was necessary to
# fully understand the data you were looking at.



#-----------------------------------------------------------------------------#
# 4.  Something Extra: Working with Animations
#-----------------------------------------------------------------------------#


# Yesterday we talked briefly about illustrating the development of a network over time using animation,
# rather than just printing network graphs side by side.

# Doing this requires installing the Image Magick software package on your system, which can be somewhat
# complex and time-consuming, so we're not going to do any exercises that require it... but if you're
# interested, here's a really quick and simple example of how to generate that kind of animation.


# We start off by creating our graph for time seried analysis just as we did yesterday:
maga_tsa = data.frame(source = maga_tweets$retweet_screen_name, 
                      target = maga_tweets$screen_name,
                      time = maga_tweets$created_at)

summary(maga_tsa$time)

# I want to create ten frames of animation, so I calculate the timestamps for each decile:
time_cutoffs <- quantile(maga_tsa$time, probs = c(0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0))

# Simplify the graph (again, as we did yesterday):
maga_tsa_net <- graph_from_data_frame(maga_tsa)
maga_tsa_net <- simplify(maga_tsa_net, remove.loops = TRUE, remove.multiple = FALSE)
decompose(maga_tsa_net, max.comps = 5)
maga_tsa_net <- decompose(maga_tsa_net)[[2]]

# Okay, from here on out we're going to do everything in a loop (which we'll pass through once per
# frame), so first of all we need to establish the layout we're going to use. Remember, it's vital
# to keep the layout consistent between frames.
set.seed(500)
l <- layout_with_fr(maga_tsa_net)

# In order to create our animation, we're going to first save every single frame as a PNG image file.
# Later we'll use Image Magick to compile these into an animated GIF file.

# The loop below is doing exactly the same thing we did yesterday in class, more or less. The only 
# real difference is that it's doing what you did in your homework assignment - looking at individual
# time slices, rather than building up the graph cumulatively.
  
  
png(file="maga_graph%02d.png", width=800, height=800)
for (i in c(1:10)) {
  
  maga_tsa_net_sub <- delete_edges(maga_tsa_net, E(maga_tsa_net)[time < time_cutoffs[[i]]])
  maga_tsa_net_sub <- delete_edges(maga_tsa_net_sub, E(maga_tsa_net_sub)[time > time_cutoffs[[i+1]]])
  
  V(maga_tsa_net_sub)$shape <- ifelse(degree(maga_tsa_net_sub) < 2,
                                    "none",
                                    "circle")
  V(maga_tsa_net_sub)$size <- ifelse(degree(maga_tsa_net_sub) < 2,
                                   0,
                                   log(degree(maga_tsa_net_sub)))
  
  plot(maga_tsa_net_sub, vertex.label = NA, vertex.frame.color = "Lightgrey", 
       edge.arrow.mode = 0, edge.color = adjustcolor("SkyBlue", alpha.f = .3), layout = l)
}
dev.off()

# Finally this command will create the animated gif file...
system("convert -delay 80 *.png maga_graph.gif")

# And this one will remove the temporary PNG files we used for each frame.
file.remove(list.files(pattern=".png"))

